/// <reference types="node" />
import http from 'http';
import EventEmitter from 'eventemitter3';
import { Logger } from 'ts-log';
import EndpointOptions from './options';
import { HandlerChain, HandlerFunction, RequestContext } from './util/handler';
import * as Handlers from './handlers';
import Session from './session';
import OcppAction from '../types/ocpp/action';
import { InboundMessage, OutboundMessage } from './message';
import { InboundCall, OutboundCall } from './call';
import { CallPayload, CallResponsePayload } from '../types/ocpp/util';
import AuthHandler, { AuthRequest } from './auth';
import { InboundMessageHandler, OutboundMessageHandler } from './handler';
interface Endpoint extends EventEmitter<EndpointEvents> {
    options: EndpointOptions;
    sessions: Map<string, Session>;
    isListening: boolean;
    listen(): void;
    stop(): void;
    drop(clientId: string, force: boolean): void;
    isConnected(clientId: string): boolean;
    handle<TRequest extends InboundCall>(action: OcppAction, callback: (data: CallPayload<TRequest>) => Promise<CallResponsePayload<TRequest>>): void;
    send<TRequest extends OutboundCall>(recipient: string, action: OcppAction, data: CallPayload<TRequest>, id?: string): Promise<CallResponsePayload<TRequest>>;
}
type EndpointEvents = {
    server_starting: (options: EndpointOptions) => void;
    server_listening: (options: EndpointOptions) => void;
    server_stopping: () => void;
    server_stopped: () => void;
    client_connecting: (request: AuthRequest) => void;
    client_accepted: (request: AuthRequest) => void;
    client_rejected: (request: AuthRequest) => void;
    client_connected: (session: Session) => void;
    client_disconnected: (session: Session) => void;
    message_sent: (message: OutboundMessage) => void;
    message_received: (message: InboundMessage) => void;
};
declare abstract class BaseEndpoint extends EventEmitter<EndpointEvents> implements Endpoint {
    readonly options: EndpointOptions;
    readonly sessions: Map<string, Session>;
    protected logger: Logger;
    protected httpServer: http.Server;
    protected authHandlers: HandlerChain<AuthHandler>;
    protected inboundHandlers: HandlerChain<InboundMessageHandler>;
    protected outboundHandlers: HandlerChain<OutboundMessageHandler>;
    abstract isConnected(clientId: string): boolean;
    abstract drop(clientId: string, force: boolean): void;
    protected abstract handleSend: HandlerFunction<OutboundMessage>;
    constructor(options: EndpointOptions, authHandlers?: AuthHandler[], inboundHandlers?: InboundMessageHandler[], outboundHandlers?: OutboundMessageHandler[], logger?: Logger, httpServer?: http.Server);
    get isListening(): boolean;
    listen(): void;
    stop(): void;
    auth(callback: Handlers.BasicAuthCallback): void;
    handle<TRequest extends InboundCall>(action: OcppAction, callback: (data: CallPayload<TRequest>) => Promise<CallResponsePayload<TRequest>>): void;
    send<TRequest extends OutboundCall>(recipient: string, action: OcppAction, data: CallPayload<TRequest>, id?: string): Promise<CallResponsePayload<TRequest>>;
    protected onSend(message: OutboundMessage): Promise<void>;
    protected onHttpError: (err: Error) => void;
    protected onAuthRequest(request: AuthRequest): void;
    protected onAuthFailure(request: AuthRequest): void;
    protected onAuthSuccess(request: AuthRequest): void;
    protected onSessionClosed(clientId: string): Promise<void>;
    protected onInboundMessage(message: InboundMessage): Promise<void>;
    protected get context(): RequestContext;
}
export default BaseEndpoint;
export { Endpoint, EndpointEvents };
