"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.OutboundMessageSynchronicityHandler = exports.InboundMessageSynchronicityHandler = void 0;
const common_tags_1 = require("common-tags");
const type_1 = __importDefault(require("../../types/ocpp/type"));
const handler_1 = require("../handler");
const call_1 = require("../call");
const callerror_1 = require("../callerror");
class InboundMessageSynchronicityHandler extends handler_1.InboundMessageHandler {
    constructor(sessions, logger) {
        super();
        this.sessions = sessions;
        this.logger = logger;
    }
    async handle(message) {
        const session = this.sessions.get(message.sender.id);
        let error = false;
        /*
        If the endpoint has sent an outbound CALL message, the client must respond
        with a CALLRESULT or CALLERROR with the same id as the outbound CALL.
        An inbound CALL from the client is possible even if no response for the
        pending outbound CALL message has been received yet.
        */
        if (session.pendingOutboundMessage &&
            !(message instanceof call_1.InboundCall) &&
            message.id !== session.pendingOutboundMessage.id) {
            error = true;
            this.logger.warn((0, common_tags_1.oneLine) `Received ${type_1.default[message.type]} message
        which is out of sync with pending outbound CALL message`);
        }
        /*
        Inbound CALLRESULT & CALLERROR messages from the client are only allowed
        if the endpoint has sent an outbound CALL message before.
        */
        if (!session.pendingOutboundMessage && !(message instanceof call_1.InboundCall)) {
            error = true;
            this.logger.warn((0, common_tags_1.oneLine) `Received ${type_1.default[message.type]} message
        while there is no pending outbound CALL message`);
        }
        /*
        The client must not send further CALL messages until the previous one
        has been responded to by the endpoint.
        */
        if (session.pendingInboundMessage && message instanceof call_1.InboundCall) {
            error = true;
            this.logger.warn((0, common_tags_1.oneLine) `Received CALL message while there
        is already a pending inbound CALL message`);
        }
        if (error) {
            throw new callerror_1.OutboundCallError(message.sender, message.id, 'ProtocolError', 'Message is out of sync', message);
        }
        return await super.handle(message);
    }
}
exports.InboundMessageSynchronicityHandler = InboundMessageSynchronicityHandler;
class OutboundMessageSynchronicityHandler extends handler_1.OutboundMessageHandler {
    constructor(sessions, logger) {
        super();
        this.sessions = sessions;
        this.logger = logger;
    }
    async handle(message) {
        const session = this.sessions.get(message.recipient.id);
        /*
        If the client has sent an inbound CALL message, the endpoint must respond
        with a CALLRESULT or CALLERROR with the same id as the inbound CALL.
        An outbound CALL from the endpoint is possible even if no response for the
        pending inbound CALL message has been sent yet.
        */
        if (session.pendingInboundMessage &&
            !(message instanceof call_1.OutboundCall) &&
            message.id !== session.pendingInboundMessage.id) {
            this.logger.warn((0, common_tags_1.oneLine) `Attempting to send ${type_1.default[message.type]} message
        which is out of sync with pending inbound CALL message`);
        }
        /*
        Outbound CALLRESULT & CALLERROR messages from the endpoint are only allowed
        if the client has sent an inbound CALL message before.
        */
        if (!session.pendingInboundMessage && !(message instanceof call_1.OutboundCall)) {
            this.logger.warn((0, common_tags_1.oneLine) `Attempting to send ${type_1.default[message.type]} message
        while there is no pending inbound CALL message`);
        }
        /*
        The endpoint must not send further CALL messages until the previous one
        has been responded to by the client.
        */
        if (session.pendingOutboundMessage && message instanceof call_1.OutboundCall) {
            this.logger.warn((0, common_tags_1.oneLine) `Attempting to send CALL message while there
        is already a pending outbound CALL message`);
        }
        return await super.handle(message);
    }
}
exports.OutboundMessageSynchronicityHandler = OutboundMessageSynchronicityHandler;
//# sourceMappingURL=message-sync.js.map