"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = __importDefault(require("path"));
const fs_1 = require("fs");
const jsonschema_1 = require("jsonschema");
const common_tags_1 = require("common-tags");
const logger_1 = __importDefault(require("../common/util/logger"));
const type_1 = __importDefault(require("../types/ocpp/type"));
const schemaBase = path_1.default.join(__dirname, '../../../var/jsonschema');
const defaultOptions = {
    schemaDir: new Map([
        // OCPP <= 1.6    /var/jsonschema/ocpp1.6
        [['ocpp1.2', 'ocpp1.5', 'ocpp1.6'], path_1.default.join(schemaBase, 'ocpp1.6')],
        // OCPP >= 2.0    /var/jsonschema/ocpp2.0.1
        [['ocpp2.0', 'ocpp2.0.1'], path_1.default.join(schemaBase, 'ocpp2.0.1')],
    ]),
};
class WsValidator {
    constructor(options = defaultOptions, logger = logger_1.default) {
        this.options = options;
        this.logger = logger;
        this.validator = new jsonschema_1.Validator();
        this.requestSchemas = new Map();
        this.responseSchemas = new Map();
    }
    async validate(type, action, data, protocol) {
        let schema;
        let schemaType;
        let schemaMap;
        switch (type) {
            case type_1.default.CALL:
                schemaType = 'request';
                schemaMap = this.requestSchemas;
                break;
            case type_1.default.CALLRESULT:
                schemaType = 'response';
                schemaMap = this.responseSchemas;
                break;
        }
        if (!schemaMap.has(action)) {
            schema = await this.loadSchema(schemaType, action, protocol);
            schemaMap.set(action, schema);
        }
        else {
            schema = schemaMap.get(action);
        }
        const result = this.validator.validate(data, schema);
        if (!result.valid) {
            this.logger.warn(`${action} ${schemaType} payload validation failed`);
            this.logger.trace(result.errors);
        }
        else {
            this.logger.debug(`${action} ${schemaType} payload validation successful`);
        }
        return result;
    }
    async loadSchema(type, action, protocol) {
        let schemaDir;
        this.options.schemaDir.forEach((dir, protocols) => {
            if (protocols.includes(protocol)) {
                schemaDir = dir;
            }
        });
        if (!schemaDir) {
            const message = (0, common_tags_1.oneLine) `Missing schema directory for
        protocol ${protocol} in WsValidator configuration`;
            this.logger.error(message);
            throw new Error(message);
        }
        const schemaPath = path_1.default.join(schemaDir, `${action}R${type.slice(1)}.json`);
        let rawSchema;
        try {
            rawSchema = await (await fs_1.promises.readFile(schemaPath)).toString();
        }
        catch (err) {
            const message = (0, common_tags_1.oneLine) `Error while attempting
        to read JSON schema from file: ${schemaPath}`;
            this.logger.error(message);
            this.logger.trace(err);
            throw new Error(message, { cause: err });
        }
        let jsonSchema;
        try {
            jsonSchema = JSON.parse(rawSchema);
        }
        catch (err) {
            const message = (0, common_tags_1.oneLine) `Error while attempting to
      parse JSON schema loaded from file: ${schemaPath}`;
            this.logger.error(message);
            this.logger.trace(err);
            throw new Error(message, { cause: err });
        }
        this.logger.debug(`Cached ${type} schema for ${action} (${protocol})`);
        return jsonSchema;
    }
}
exports.default = WsValidator;
//# sourceMappingURL=ws-validator.js.map